"""Common module"""

import os

import numpy as np


def scidx(bw, ng):
    """subcarriers index

    Args:
        bw: bandwitdh(20, 40)
        ng: grouping(1, 2, 4)
    """
    if bw not in [20, 40] or ng not in [1, 2, 4]:
        raise ValueError("bw should be [20, 40] and ng should be [1, 2, 4]")
    a, b = int(bw * 1.5 - 2), int(bw / 20)
    k = np.r_[range(-a, -b, ng), -b, range(b, a, ng), a]
    return k


def calib(phase, bw=20, ng=2, axis=1):
    """Phase calibration

    Note:
        phase: it must be unwrapped, it should be a 2-D, 3-D or 4-D array and
            the second dimension must be subcarriers
        bw, ng: the same as `scidx`

    ref:
        [Enabling Contactless Detection of Moving Humans with Dynamic Speeds Using CSI]
        (http://tns.thss.tsinghua.edu.cn/wifiradar/papers/QianKun-TECS2017.pdf)
    """
    p = np.asarray(phase)
    s = scidx(bw, ng)

    slice1 = [slice(None, None)] * p.ndim
    slice1[axis] = slice(-1, None)
    slice1 = tuple(slice1)
    slice2 = [slice(None, None)] * p.ndim
    slice2[axis] = slice(None, 1)
    slice2 = tuple(slice2)
    shape1 = [1] * p.ndim
    shape1[axis] = s.shape[0]
    shape1 = tuple(shape1)

    k_n, k_1 = s[-1], s[1]
    a = (p[slice1] - p[slice2]) / (k_n - k_1)
    b = p.mean(axis=axis, keepdims=True)
    s = s.reshape(shape1)

    phase_calib = p - a * s - b
    return phase_calib


def phy_ifft(x, bw=20, ng=2, axis=1):
    """802.11n IFFT

    Return discrete inverse Fourier transform of real or complex sequence. It
    is based on Equation (19-25)(P2373)

    Note:
        1. No ifftshift
        2. scipy.fftpack.ifft is different from Equation (19-25) and Equation (17-9)
        3. BE CAREFUL! I haven't found any code about CSI like this.

    Ref:
        IEEE Standard for Information technology—Telecommunications and information
        exchange between systems Local and metropolitan area networks—Specific 
        requirements - Part 11: Wireless LAN Medium Access Control (MAC) and Physical
        Layer (PHY) Specifications, in IEEE Std 802.11-2016 (Revision of IEEE Std
        802.11-2012), vol., no., pp.1-3534, 14 Dec. 2016, doi: 10.1109/IEEESTD.2016.7786995.
    """
    assert bw == 20, "Only bw=20 is allowed"

    x = np.expand_dims(x.swapaxes(-1, axis), -2)
    k = scidx(bw, ng)

    n = 64 * (bw / 20)
    delta_f = bw * 1e6 / n
    t = np.arange(n).reshape(-1, 1) / (bw * 1e6)

    out = (x * np.exp(1.j * 2 * np.pi * k * delta_f * t)).mean(axis=-1).swapaxes(-1, axis)
    return out


def phy_fft(x, bw=20, ng=2, axis=1):
    """802.11n FFT

    Return discrete Fourier transform of real or complex sequence.
    """
    assert bw == 20, "Only bw=20 is allowed"

    x = np.expand_dims(x.swapaxes(-1, axis), -1)
    k = scidx(bw, ng)

    n = 64 * (bw / 20)
    delta_f = bw * 1e6 / n
    t = np.arange(n).reshape(-1, 1) / (bw * 1e6)

    scale = k.size / n
    out = (x * np.exp(1.j * -2 * np.pi * k * delta_f * t)).sum(axis=-2).swapaxes(-1, axis) * scale
    return out


def infer_device(csifile):
    """Infer the CSI file format simplely

    Args:
        csifile (str): csi data file

    Returns:
        str:
            Intel: intel 5300 csi file
            Atheros: atheros csi file
            Nexmon: nexmon csi file
            AtherosPull10: `atheros pull 10 <https://github.com/xieyaxiongfly/Atheros-CSI-Tool-UserSpace-APP/pull/10/files>`_
            NexmonPull46: `nexmon_csi pull 46 <https://github.com/seemoo-lab/nexmon_csi/pull/46/files>`_
            Unknown: Unknown file format
    Note:
        This function cannot work for some nexmon csi file formats defined by
        projects derived from nexmon_csi
    """
    with open(csifile, 'rb') as f:
        buf = f.read(4)
        if buf[2] in [0xc1, 0xbb]:
            return 'Intel'
        elif buf in [b"\xa1\xb2\xc3\xd4", b"\xd4\xc3\xb2\xa1",
                     b"\xa1\xb2\x3c\x4d", b"\x4d\x3c\xb2\xa1"]:
            f.seek(20 + 16 + 42, os.SEEK_CUR)
            buf = f.read(4)
            if buf == b'\x11\x11\x11\x11':
                return 'Nexmon'
            if buf[:2] == b'\x11\x11' and buf[2:] != b'\x11\x11':
                return 'NexmonPull46'
        elif buf[0] in [0xff, 0x00]:
            return 'AtherosPull10'
        else:
            f.seek(2 + 16 - 4, os.SEEK_CUR)
            buf = f.read(1)
            if buf[0] == 56 or buf[0] == 114:
                return 'Atheros'
            else:
                return 'Unknown'


def infer_tones(csifile, device):
    """Infer the argument `tones` of Atheros and AtherosPull10
    
    Args:
        csifile (str): atheros csi file
        device (str): Atheros or AtherosPull10
    
    Returns:
        int: tones
    
    Examples:

        >>> tones = infer_tones(csifile, 'Atheros')
        >>> csidata = csiread.Atheros(csifile, nrxnum=3, ntxnum=3,
        >>>                           if_report=False, tones=tones)
    """
    with open(csifile, 'rb') as f:
        if device == "Atheros":
            f.seek(18, os.SEEK_CUR)
        elif device == 'AtherosPull10':
            f.seek(19, os.SEEK_CUR)
        else:
            raise Exception("It is not a Atheros file")
        tones = f.read(1)[0]
    return tones


def infer_chip_bw(csifile):
    """Infer chip, bandwidth and band of nexmon csi file.

    This function may be failed on some nexmon csi files.

    Args:
        csifile (str): nexmon csi file

    Returns:
        tuple (chip, bw, band):
            str: chip
            int: bw, MHz
            int: band, GHz 

    Examples:

        >>> csifile = "../material/nexmon/dataset/example.pcap"
        >>> chip, bw, band = nexchan(csifile)
        >>> csidata = csiread.Nexmon(csifile, chip=chip, bw=bw)
        >>> csidata.read()

    References:
        1. `D11AC_IOTYPES lines 172-188 <https://github.com/seemoo-lab/nexmon/blob/master/patches/include/bcmwifi_channels.h#L172>`_
    """
    csidata = csiread.Nexmon(None, 'none', 0, False, bufsize=1)
    csidata.seek(csifile, 24, 1)
    chan_spec = csidata.chan_spec[0]
    chip_version = csidata.chip_version[0]

    CHIP_LIST = {0x0065: '43455c0', 0xdead: '4358', 0xe834: '4366c0'}
    WL_CHANSPEC_BW = chan_spec & 0x3800
    WL_CHANSPEC_BAND = chan_spec & 0xc000
    if WL_CHANSPEC_BW in [0x2800, 0x3000]:
        raise Exception("csiread.Nexmon doesn't support WL_CHANSPEC_BW_160 "
                        "and WL_CHANSPEC_BW_8080")
    if WL_CHANSPEC_BAND == 255:
        raise Exception("INVCHANSPEC")

    chip = CHIP_LIST[chip_version]
    bw = (1 << (WL_CHANSPEC_BW // 0x0800)) * 5
    band = WL_CHANSPEC_BAND // 0x4000 + 2

    return chip, bw, band



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI client: Create a large csi data file for testing csiread's performance(Linux 802.11n CSI Tool)

Usage:
    1. python3 csiclient.py
    2. python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 1000
"""

import socket


def create_large_csifile(csifile='csifile.dat', endian='big'):
    """Create a large csi data file for testing csiread's performance
    
    Note:
        5300NIC: 'endian' must be 'big'
        Atheros: 'endian' shoud be the same as csiserver.py's
    """

    # config
    address_src = ('127.0.0.1', 10086)
    address_des = ('127.0.0.1', 10010)

    count = 0
    f = open(csifile, 'wb')
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.bind(address_des)
        while True:
            data, address_src = s.recvfrom(4096)
            msg_len = len(data)
            f.write(msg_len.to_bytes(2, endian))
            f.write(data)
            count += 1
            print('received %d bytes, id: %d, seq: %d, clen: %d' % (msg_len, 0, 0, count))
            if count % 100 == 0:
                print('wrote %d bytes [msgcnt=%u]' % (msg_len, count))
    f.close()


if __name__ == "__main__":
    create_large_csifile(csifile='csiclient.dat', endian='big')
    # create_large_csifile(csifile='csiclient.dat', endian='little')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""MUSIC Algorithm

Usage:
    python3 csimusic.py

Ref:
   1. [MUSIC Super-Resolution DOA Estimation](https://ww2.mathworks.cn/help/phased/ug/music-super-resolution-doa-estimation.html)
   2. [Phased Array System Toolbox](https://ww2.mathworks.cn/en/products/phased-array.html)
   3. 王永良. 空间谱估计理论与算法[M]. 清华大学出版社, 2004.
"""

import numpy as np
from numpy import linalg as LA
from scipy.signal import find_peaks
import matplotlib.pyplot as plt


def signalG(N, M, L, d, f, c=3e8):
    """Signal Generation(Uncorrelated Sources)

    Args:
        N: number of snapshots
        M: number of receivers
        L: number of sources
        d: distance between receivers
        f: frequency
        c: wave speed
    """
    ula = np.c_[:M] * d                                         # uniform linear array (ULA)
    doa = np.sort(np.random.uniform(-np.pi/2, np.pi/2, L))      # directions of arrival (DOA)
    a = np.exp(-2.j * np.pi * f * ula * np.sin(doa) / c)        # steering vector
    s = np.random.randn(L, N) + np.random.randn(L, N) * 1.j     # source signal
    n = np.random.randn(M, N) + np.random.randn(M, N) * 1.j     # noise
    x = a @ s + n * 0.01                                        # signal model
    return x.T, doa


def NoSS(X, method='AIC'):
    """Number of Signal Source"""
    N, M = X.shape
    R = (X[..., np.newaxis] @ X[:, np.newaxis].conj()).mean(axis=0)
    w, v = LA.eigh(R)

    def Lambda(n):
        return 1 / (M - n) * np.sum(w[:-n]) / np.prod(w[:-n]) ** (1 / (M - n))

    def AIC(n):
        return 2 * N * (M - n) * np.log(Lambda(n)) + 2 * n * (2 * M - n)

    def MDL(n):
        return N * (M - n) * np.log(Lambda(n)) + 1/2 * n * (2 * M - n) * np.log(N)

    def HQ(n):
        return N * (M - n) * np.log(Lambda(n)) + 1/2 * n * (2 * M - n) * np.log(np.log(N))

    func = eval(method)
    L = np.argmin([func(i) for i in range(1, M)]) + 1
    return L


def music(X, L, d, f, c=3e8):
    """MUSIC Algorithm

    Args:
        X: signal received [snapshots, receivers]
        L: number of sources
        d: distance between receivers
        f: frequency
        c: wave speed
    """
    N, M = X.shape
    R = (X[..., np.newaxis] @ X[:, np.newaxis].conj()).mean(axis=0)
    w, v = LA.eigh(R)
    u = v[:, ::-1][:, L:]                                       # noise subspace

    ula = np.c_[:M] * d                                         # ULA  
    doa_space = np.linspace([[-np.pi/2]], [[np.pi/2]], 180)     # search space
    a = np.exp(-2.j * np.pi * f * ula * np.sin(doa_space) / c)  # steering vector
    p = 1 / LA.norm(u.T.conj() @ a, axis=(1, 2))                # MUSIC pseudospectrum
    # p *= LA.norm(a, axis=(1, 2))                                # optional
    p = 10 * np.log10(p/p.max())

    # DOA
    peak, _ = find_peaks(p)
    index = sorted(peak[p[peak].argsort()][-L:])
    doa = doa_space[index, 0, 0]
    return p, doa


def music_root(X, L, d, f, c=3e8):
    """Root-MUSIC(RMU) Algorithm"""
    N, M = X.shape
    R = (X[..., np.newaxis] @ X[:, np.newaxis].conj()).mean(axis=0)
    w, v = LA.eigh(R)
    u = v[:, ::-1][:, L:]

    coff = np.r_[1, (u[:L] @ LA.inv(u[L:]))[::-1, 0]]
    r = np.roots(coff)
    doa = np.sort(np.arcsin(np.angle(r) * c / f / (2 * np.pi * d)))
    return None, doa


def plotspectrum(p, doa_fake, doa_real):
    doa_real *= (180 / np.pi)
    doa_fake *= (180 / np.pi)
    print("signalG - doa: ", doa_real)
    print("MUSIC   - doa: ", doa_fake)
    
    if p is None:
        p = np.sin(np.deg2rad(np.arange(180)))
    x = np.linspace(-90, 90, p.size)
    index = [np.abs(x - a).argmin() for a in doa_fake]

    plt.figure()
    plt.plot(x, p)
    plt.plot(doa_fake, p[index], 'xr', markersize=10, label='doa fake')
    plt.vlines(doa_real, p.min(), p.max(), colors='orange', linewidth=2,
               linestyles='--', label='doa real')
    plt.legend()
    plt.title("MUSIC")
    plt.xlabel("Angle(deg)")
    plt.ylabel("Power(dB)" if p.max() != 1 else "")
    plt.show()


if __name__ == "__main__":
    x, doa_real = signalG(N=100, M=4, L=3, d=6.25e-2, f=2.4e9)
    p, doa_fake = music(X=x, L=NoSS(x), d=6.25e-2, f=2.4e9)

    plotspectrum(p, doa_fake, doa_real)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Plot CSI(Linux 802.11n CSI Tool) in real time

Usage:
    1. python3 csirealtime.py
    2. python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 500
"""

import socket
import threading

import csiread
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np

cache_len = 1000
cache_data1 = [np.nan] * cache_len
cache_data2 = [np.nan] * cache_len
cache_data3 = [np.nan] * cache_len
mutex = threading.Lock()


class GetDataThread(threading.Thread):
    def run(self):
        update_background()


def update_background():
    csidata = csiread.Intel(None, 3, 1)

    # config
    global cache_data1, cache_data2, cache_data3, mutex
    count = 0
    address_src = ('127.0.0.1', 10086)
    address_des = ('127.0.0.1', 10010)

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.bind(address_des)
        while True:
            data, address_src = s.recvfrom(4096)
            msg_len = len(data)

            code = csidata.pmsg(data)
            if code == 0xbb:
                scaled_csi_sm = csidata.get_scaled_csi()[0]
                mutex.acquire()
                cache_data1.pop(0)
                cache_data1.append(scaled_csi_sm[15, 0, 0])
                cache_data2.pop(0)
                cache_data2.append(scaled_csi_sm[15, 1, 0])
                cache_data3.pop(0)
                cache_data3.append(scaled_csi_sm[15, 2, 0])
                mutex.release()
                count += 1
            if count % 100 == 0:
                print('receive %d bytes [msgcnt=%u]' % (msg_len, count))


def realtime_plot():
    fig, ax = plt.subplots()
    plt.title('csi-amplitude')
    plt.xlabel('packets')
    plt.ylabel('amplitude')
    ax.set_ylim(0, 40)
    ax.set_xlim(0, cache_len)
    x = np.arange(0, cache_len, 1)

    line1,  = ax.plot(x, np.abs(cache_data1), linewidth=1.0, label='subcarrier_15_0_0')
    line2,  = ax.plot(x, np.abs(cache_data2), linewidth=1.0, label='subcarrier_15_1_0')
    line3,  = ax.plot(x, np.abs(cache_data3), linewidth=1.0, label='subcarrier_15_2_0')
    plt.legend()

    def init():
        line1.set_ydata([np.nan] * len(x))
        line2.set_ydata([np.nan] * len(x))
        line3.set_ydata([np.nan] * len(x))
        return line1, line2, line3,

    def animate(i):
        global cache_data1, cache_data2, cache_data3, mutex
        mutex.acquire()
        line1.set_ydata(np.abs(cache_data1))
        line2.set_ydata(np.abs(cache_data2))
        line3.set_ydata(np.abs(cache_data3))
        mutex.release()
        return line1, line2, line3,

    ani = animation.FuncAnimation(fig, animate, init_func=init, interval=1000/25, blit=True)
    plt.show()


if __name__ == '__main__':
    task = GetDataThread()
    task.start()
    realtime_plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""csiread.Nexmon: implemented in pure Python

Note:
    csiread.Nexmon is based on the commit of nexmon_csi(Aug 29, 2020):
    ba99ce12a6a42d7e4ec75e6f8ace8f610ed2eb60

Ref:
    [nexmon_csi](https://github.com/seemoo-lab/nexmon_csi)
    [rdpcap](https://github.com/secdev/scapy/blob/master/scapy/utils.py)
"""

import os
import struct
import numpy as np
from timeit import default_timer


def unpack_int16(buf, csi, nfft, endian):
    dt = np.dtype(np.int16).newbyteorder(endian)
    csi[:] = np.frombuffer(buf, dtype=dt, count=nfft)


def unpack_float(buf, csi, nfft, M, E, endian):
    nbits = 10
    autoscale = 1
    e_p = (1 << (E - 1))
    e_shift = 1
    e_zero = - M
    maxbit = -e_p
    k_tof_unpack_sgn_mask = (1 << 31)
    ri_mask = (1 << (M - 1)) - 1
    E_mask = (1 << E) - 1
    sgnr_mask = (1 << (E + 2 * M - 1))
    sgni_mask = sgnr_mask >> M
    He = [0] * 256
    Hout = [0] * 512
    border = 'little' if endian == '<' else 'big'

    for i in range(nfft):
        h = int.from_bytes(buf[4*i:4*i+4], border)

        v_real = (h >> (E + M)) & ri_mask
        v_imag = (h >> E) & ri_mask
        e = h & E_mask
        if (e >= e_p):
            e -= (e_p << 1)
        He[i] = e
        x = v_real | v_imag

        if (autoscale and x):
            m = 0xffff0000
            b = 0xffff
            s = 16
            while (s > 0):
                if (x & m):
                    e += s
                    x >>= s
                s >>= 1
                m = (m >> s) & b
                b >>= s
            if (e > maxbit):
                maxbit = e
        if (h & sgnr_mask):
            v_real |= k_tof_unpack_sgn_mask
        if (h & sgni_mask):
            v_imag |= k_tof_unpack_sgn_mask
        Hout[i<<1] = v_real
        Hout[(i<<1)+1] = v_imag

    shft = nbits - maxbit
    for i in range(nfft*2):
        e = He[(i >> e_shift)] + shft
        sgn = 1
        if (Hout[i] & k_tof_unpack_sgn_mask):
            sgn = -1
            Hout[i] &= ~k_tof_unpack_sgn_mask
        if (e < e_zero):
            Hout[i] = 0
        elif (e < 0):
            e = -e
            Hout[i] = (Hout[i] >> e)
        else:
            Hout[i] = (Hout[i] << e)
        Hout[i] *= sgn

    for i in range(nfft):
        csi[i] = Hout[i*2] + Hout[i*2+1] * 1.j


class Nexmon:
    def __init__(self, file, chip, bw, if_report=True):
        self.file = file
        self.chip = chip
        self.bw = bw
        self.if_report = if_report

        if file is None:
            self.count = 0
            pk_num = 1
        else:
            pk_num = self.__get_count()

        btype = np.int_
        self.sec = np.zeros([pk_num], dtype=btype)
        self.usec = np.zeros([pk_num], dtype=btype)
        self.caplen = np.zeros([pk_num], dtype=btype)
        self.wirelen = np.zeros([pk_num], dtype=btype)
        self.magic = np.zeros([pk_num], dtype=btype)
        self.src_addr = np.zeros([pk_num, 6], dtype=btype)
        self.seq = np.zeros([pk_num], dtype=btype)
        self.core = np.zeros([pk_num], dtype=btype)
        self.spatial = np.zeros([pk_num], dtype=btype)
        self.chan_spec = np.zeros([pk_num], dtype=btype)
        self.chip_version = np.zeros([pk_num], dtype=btype)
        self.csi = np.zeros([pk_num, int(self.bw * 3.2)], dtype=np.complex_)

    def read(self):
        count = 0
        nfft = int(self.bw * 3.2)

        f = open(self.file, 'rb')
        endian = self.__pcapheader(f)

        while True:
            hdr = f.read(16)
            if len(hdr) < 16:
                break
            sec, usec, caplen, wirelen = struct.unpack(endian + "IIII", hdr)
            buf = f.read(42)
            if buf[6:12] != b'NEXMON':
                f.seek(caplen - 42, os.SEEK_CUR)
                continue
            self.sec[count] = sec
            self.usec[count] = usec
            self.caplen[count] = caplen
            self.wirelen[count] = wirelen

            frame = f.read(18)
            magic, src_addr, seq, core_spatial, chan_spec, chip_version = struct.unpack(
                endian + "I6sHHHH", frame
            )
            
            self.magic[count] = magic
            for i in range(6):
                self.src_addr[count, i] = src_addr[i]
            self.seq[count] = seq
            self.core[count] = core_spatial & 0x7
            self.spatial[count] = (core_spatial >> 3) & 0x7
            self.chan_spec[count] = chan_spec
            self.chip_version[count] = chip_version

            buf = f.read(caplen - 42 - 18)
            if self.chip == '4339' or self.chip == '3455c0':
                unpack_int16(buf, self.csi[count], nfft, endian)
            elif self.chip == '4358':
                unpack_float(buf, self.csi[count], nfft, 9, 5, endian)
            elif self.chip == '4366c0':
                unpack_float(buf, self.csi[count], nfft, 12, 6, endian)
            else:
                pass
            count += 1
        f.close
        self.count = count
        if self.if_report:
            print(str(count) + " packets parsed")

    def __get_count(self):
        count = 0
        f = open(self.file, 'rb')
        endian = self.__pcapheader(f)
        while True:
            hdr = f.read(16+42)
            if len(hdr) < 16:
                break
            if hdr[22:28] == b'NEXMON':
                count += 1
            caplen, = struct.unpack(endian + "I", hdr[8:12])
            f.seek(caplen - 42, os.SEEK_CUR)
        f.close()
        return count

    def __pcapheader(self, f):
        magic = f.read(4)
        if magic == b"\xa1\xb2\xc3\xd4":  # big endian
            endian = ">"
            self.nano = False
        elif magic == b"\xd4\xc3\xb2\xa1":  # little endian
            endian = "<"
            self.nano = False
        elif magic == b"\xa1\xb2\x3c\x4d":  # big endian, nanosecond-precision
            endian = ">"
            self.nano = True
        elif magic == b"\x4d\x3c\xb2\xa1":  # little endian, nanosecond-precision
            endian = "<"
            self.nano = True
        else:
            raise Exception("Not a pcap capture file (bad magic: %r)" % magic)

        f.seek(20, os.SEEK_CUR)
        return endian


if __name__ == '__main__':
    last = default_timer()
    csifile = "../material/nexmon/dataset/example.pcap"
    csidata = Nexmon(csifile, chip='4358', bw=80, if_report=False)
    csidata.read()
    print(default_timer() - last, 's')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""SpotFi: `SpotFi: Decimeter Level Localization Using WiFi`

Usage:
    python3 csispotfi.py

Important:
    Unfinished! There are many issues.

Ref:
    1. [SpotFi: Decimeter Level Localization Using WiFi](http://conferences.sigcomm.org/sigcomm/2015/pdf/papers/p269.pdf)
    2. [spotfiMusicAoaEstimation](https://bitbucket.org/mkotaru/spotfimusicaoaestimation)
"""

import numpy as np
from numpy import linalg as LA
from scipy.signal import find_peaks
import csiread
import matplotlib.pyplot as plt
from utils import scidx


def remove_sto(csi, bw=20, ng=2):
    """Algorithm 1: SpotFi’s ToF sanitization algorithm

    Args:
        csi: [30, 3] - [subcarriers, Ntx]

    Ref:
        spotfiMusicAoaEstimation: removePhsSlope.m
    """
    s_index = np.tile(scidx(bw, ng), (csi.shape[1], 1)).T

    # In the paper, step 1 is least-squares 
    m, c = np.linalg.lstsq(np.c_[s_index.flatten('F'), np.ones(csi.size)], 
                           np.unwrap(np.angle(csi), axis=0).flatten('F'),
                           rcond=None)[0]
    return csi * np.exp(-1.j * (m * s_index + c))


def smooth_csi(csi):
    """Fig.4: CSI smoothing"""
    smoothed_csi = np.zeros([30, 32], dtype=np.complex_)
    for i in range(15):
        smoothed_csi[i] = csi[i:i+16, 0:2].flatten('F')
        smoothed_csi[i+15] = csi[i:i+16:, 1:3].flatten('F')
    return smoothed_csi


def smooth_csiB(csi):
    """Fig.4: CSI smoothing(another version)"""
    from scipy.linalg import hankel
    h0 = hankel(csi[:15, 0], csi[14:, 0])
    h1 = hankel(csi[:15, 1], csi[14:, 1])
    h2 = hankel(csi[:15, 2], csi[14:, 2])
    smoothed_csi = np.c_[np.r_[h0, h1],
                         np.r_[h1, h2]]
    return smoothed_csi


def spotfi(csi):
    pass



import numpy as np
from .csiread import Atheros, Nexmon


def scidx(bw, ng):
    """SubCarrier InDeX

    Table 9-54-Number of matrices and carrier grouping

    Args:
        bw (int): Bandwidth, it can be 20 and 40.
        ng (int): Grouping, it can be 1, 2 and 4.

    Returns:
        ndarray: Subcarrier index

    Examples:

        >>> s_index = scidx(20, 2)

    References:
        1. `IEEE Standard for Information technology—Telecommunications and
        information exchange between systems Local and metropolitan area
        networks—Specific requirements - Part 11: Wireless LAN Medium Access
        Control (MAC) and Physical Layer (PHY) Specifications, in
        IEEE Std 802.11-2016 (Revision of IEEE Std 802.11-2012), vol., no.,
        pp.1-3534, 14 Dec. 2016, doi: 10.1109/IEEESTD.2016.7786995. <#>`_
    """
    if bw not in [20, 40] or ng not in [1, 2, 4]:
        raise ValueError("bw should be [20, 40] and ng should be [1, 2, 4]")
    a, b = int(bw * 1.5 - 2), int(bw / 20)
    k = np.r_[-a:-b:ng, -b, b:a:ng, a]
    return k


def calib(phase, bw=20, ng=2, axis=1):
    """Phase calibration

    Args:
        phase (ndarray): Unwrapped phase of CSI.
        bw (int): Bandwidth, it can be 20 and 40. Default: 20
        ng (int): Grouping, it can be 1, 2 and 4. Default: 2
        axis (int): Axis along which is subcarrier. Default: 1

    Returns:
        ndarray: Phase calibrated

    Examples:

        >>> csi = csidata.csi[:10]
        >>> phase = np.unwrap(np.angle(csi), axis=1)
        >>> phase = calib(phase, bw=20, ng=2, axis=1)

    References:
        1. `Enabling Contactless Detection of Moving Humans with Dynamic Speeds
        Using CSI <#>`_
    """
    p = np.asarray(phase)
    s = scidx(bw, ng)

    slice1 = [slice(None, None)] * p.ndim
    slice1[axis] = slice(-1, None)
    slice1 = tuple(slice1)
    slice2 = [slice(None, None)] * p.ndim
    slice2[axis] = slice(None, 1)
    slice2 = tuple(slice2)
    shape1 = [1] * p.ndim
    shape1[axis] = s.shape[0]
    shape1 = tuple(shape1)

    k_n, k_1 = s[-1], s[1]
    a = (p[slice1] - p[slice2]) / (k_n - k_1)
    b = p.mean(axis=axis, keepdims=True)
    s = s.reshape(shape1)

    phase_calib = p - a * s - b
    return phase_calib


class AtherosPull10(Atheros):
    """Parse CSI obtained using 'Atheros CSI Tool' pull 10

    References:
        1. `Atheros-CSI-Tool-UserSpace-APP pull 10 <https://github.com/xieyaxiongfly/Atheros-CSI-Tool-UserSpace-APP/pull/10>`_
    """
    def read(self):
        with open(self.file, 'rb') as f:
            endian = 'big' if f.read(1) == b'\xff' else 'little'
        self.seek(self.file, 1, 0, endian)


class NexmonPull46(Nexmon):
    """Parse CSI obtained using 'nexmon_csi' pull 46

    Tips:
        The status code returned by pmsg() is ``0xf101``

    References:
        1. `nexmon_csi pull 46 <https://github.com/seemoo-lab/nexmon_csi/pull/46>`_
    """
    def __init__(self, file, chip, bw, if_report=True, bufsize=0):
        super(NexmonPull46, self).__init__(file, chip, bw, if_report, bufsize)
        self.rssi = None
        self.fc = None
        self._autoscale = 0

    def __getitem__(self, index):
        ret = super().__getitem__(index)
        ret['rssi'] = self.rssi[index]
        ret['fc'] = self.fc[index]
        return ret

    def seek(self, file, pos, num):
        super().seek(file, pos, num)
        self.__pull46()

    def pmsg(self, data, endian='little'):
        super().pmsg(data, endian)
        self.__pull46()
        return 0xf101

    def __pull46(self):
        if self.magic[0] & 0x0000ffff == 0x1111:
            self.rssi = ((self.magic & 0x00ff0000) >> 16)
            self.rssi = self.rssi.astype(np.int8).astype(np.int_)
            self.fc = (self.magic & 0xff000000) >> 24
            self.magic &= 0x0000ffff
        else:
            self.rssi = ((self.magic & 0x0000ff00) >> 8)
            self.rssi = self.rssi.astype(np.int8).astype(np.int_)
            self.fc = self.magic & 0x000000ff
            self.magic = (self.magic & 0xffff0000) >> 16



"""csiread.ESP32: implemented in pure Python

Note: There is no plan to implement csiread.ESP32 with Cython.

Ref: [ESP32-CSI-Tool](https://github.com/StevenMHernandez/ESP32-CSI-Tool).
    Get test data from here.
"""

from timeit import default_timer

import numpy as np
import os


class ESP32:
    def __init__(self, file, if_report=True):
        self.file = file
        self.if_report = if_report

        pk_num = self.get_count()
        self.type = np.zeros([pk_num], 'U8')
        self.role = np.zeros([pk_num], 'U8')
        self.mac = np.zeros([pk_num], 'U17')
        self.rssi = np.zeros([pk_num], np.int_)
        self.rate = np.zeros([pk_num], np.int_)
        self.sig_mode = np.zeros([pk_num], np.int_)
        self.mcs = np.zeros([pk_num], np.int_)
        self.bandwidth = np.zeros([pk_num], np.int_)
        self.smoothing = np.zeros([pk_num], np.int_)
        self.not_sounding = np.zeros([pk_num], np.int_)
        self.aggregation = np.zeros([pk_num], np.int_)
        self.stbc = np.zeros([pk_num], np.int_)
        self.fec_coding = np.zeros([pk_num], np.int_)
        self.sgi = np.zeros([pk_num], np.int_)
        self.noise_floor = np.zeros([pk_num], np.int_)
        self.apmdu_cnt = np.zeros([pk_num], np.int_)
        self.channel = np.zeros([pk_num], np.int_)
        self.secondary_channel = np.zeros([pk_num], np.int_)
        self.local_timestamp = np.zeros([pk_num], np.int_)
        self.ant = np.zeros([pk_num], np.int_)
        self.sig_len = np.zeros([pk_num], np.int_)
        self.rx_state = np.zeros([pk_num], np.int_)
        self.real_time_set = np.zeros([pk_num], np.int_)
        self.real_timestamp = np.zeros([pk_num], np.float_)
        self.len = np.zeros([pk_num], np.int_)
        self.csi_data = np.zeros([pk_num, 64], dtype=np.complex_)

    def read(self):
        with open(self.file) as f:
            count = 0
            for line in f.readlines():
                line = line.split(',')
                self.type[count] = line[0]
                self.role[count] = line[1]
                self.mac[count] = line[2]
                self.rssi[count] = int(line[3])
                self.rate[count] = int(line[4])
                self.sig_mode[count] = int(line[5])
                self.mcs[count] = int(line[6])
                self.bandwidth[count] = int(line[7])
                self.smoothing[count] = int(line[8])
                self.not_sounding[count] = int(line[9])
                self.aggregation[count] = int(line[10])
                self.stbc[count] = int(line[11])
                self.fec_coding[count] = int(line[12])
                self.sgi[count] = int(line[13])
                self.noise_floor[count] = int(line[14])
                self.apmdu_cnt[count] = int(line[15])
                self.channel[count] = int(line[16])
                self.secondary_channel[count] = int(line[17])
                self.local_timestamp[count] = int(line[18])
                self.ant[count] = int(line[19])
                self.sig_len[count] = int(line[20])
                self.rx_state[count] = int(line[21])
                self.real_time_set[count] = int(line[22])
                self.real_timestamp[count] = float(line[23])
                self.len[count] = int(line[24])
                cline = line[-1].strip('[]').split(' ')[:-1]
                csi = [int(i) + int(j) * 1.j for i, j in zip(cline[::2], cline[1::2])]
                self.csi_data[count] = csi
                count += 1

    def get_count(self):
        lens = os.path.getsize(self.file)
        with open(self.file) as f:
            linesize = len(f.readline())
        return lens // linesize

    def create_large_file(self, count):
        with open(self.file) as f:
            data = f.readline()
        with open("esp32_lagre_file.csv", 'a+') as f:
            for i in range(count):
                f.write(data)


if __name__ == '__main__':
    last = default_timer()
    csifile = '../material/esp32/dataset/example_csi.csv'
    csidata = ESP32(csifile, if_report=True)
    csidata.read()
    print(default_timer() - last, 's')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI viewer: Advanced version of csishow.py

> Matplotlib is a Python 2D plotting library which produces publication quality figures
> in a variety of hardcopy formats and interactive environments across platforms

> If you are doing anything requiring rapid plot updates, video, or realtime
> interactivity, matplotlib is not the best choice.

Usage:
    1. python3 csiviewer.py
    2. python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 2000
    3. F11: show fullscreen, F1: show menu bar.

Note:
    1. Functions such as np.angle() will take some time, so csiserver.py uses 
        a larger value of delay. downsampling would be better.
    2. If you want to start it quickly on Linux, create a file named 'csiviewer.desktop'. 
        Copy the following code to 'csiviewer.desktop', and give it executable permissions.
        Finally, you need set `Exec` manually.

        ```bash
        [Desktop Entry]
        Name=CSIViewer
        Comment=CSI Plot
        Keywords=CSI;
        StartupNotify=true
        Terminal=false
        Type=Application
        Categories=Application;
        Icon=/opt/csiread/csiviewer.png
        Exec=python3 /opt/csiread/csiviewer.py %F
        Name[en_US]=csiviewer.desktop
"""

import os
import socket
import sys

import csiread
import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import QMutex, Qt, QThread, QTimer, pyqtSlot
from PyQt5.QtGui import QApplication
from PyQt5.QtWidgets import QLabel, QMenuBar, QPushButton, QVBoxLayout, QWidget
from utils import scidx, phy_ifft, calib


os.environ['QT_SCALE_FACTOR'] = '1'

Device = 'Intel' # Intel or Atheros

if Device == 'Intel':
    csidata = csiread.Intel(None, 3, 2)
    SUBCARRIERS_NUM = 30
    BW = 20
    NG = 2
    YRange_A = 70
    YRange_B = 70
    YRange_D = 40
elif Device == 'Atheros':
    csidata = csiread.Atheros(None, 3, 2)
    SUBCARRIERS_NUM = 56
    BW = 20
    NG = 1
    YRange_A = 300
    YRange_B = 300
    YRange_D = 200
else:
    raise ValueError("Device = 'Intel' or 'Atheros'")



cache_amptiA = np.zeros([SUBCARRIERS_NUM, 800])       # [subcarriers, packets]
cache_amptiB = np.zeros([SUBCARRIERS_NUM, 10, 3])     # [subcarriers, packets, Nrx]
cache_phaseC = np.zeros([SUBCARRIERS_NUM, 10, 3])     # [subcarriers, packets, Nrx]
cache_cirD = np.zeros([64, 10, 3])       # [time(samples), packets, Nrx]

mutex = QMutex()
state = True


class GetDataThread(QThread):
    def __init__(self, parent):
        super(GetDataThread, self).__init__(parent)

    def stop(self):
        self.requestInterruption()
        self.exit()

    def run(self):
        """get data in real time

        Note:
            If you want to run this script on the host with Intel 5300 NIC, rewrite code as
            csiuserspace.py
        """

        # config
        global cache_amptiA, cache_amptiB, cache_phaseC, cache_cirD, mutex, state
        count = 0
        address_src = ('127.0.0.1', 10086)
        address_des = ('127.0.0.1', 10010)

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.bind(address_des)
            s.settimeout(1/30)
            # while True:       # `QThread: Destroyed while thread is still running`
            while not self.isInterruptionRequested():
                try:
                    data, address_src = s.recvfrom(4096)
                except socket.timeout:
                    continue
                msg_len = len(data)

                code = csidata.pmsg(data)  # Notice: endian
                if code == 0xbb:            # Intel
                    csi = csidata.get_scaled_csi_sm(True)
                elif code == 0xff00:        # Atheros
                    csi = csidata.csi
                else:
                    continue

                mutex.lock()
                cache_amptiA[:, :-1] = cache_amptiA[:, 1:]
                cache_amptiB[:, :-1] = cache_amptiB[:, 1:]
                cache_phaseC[:, :-1] = cache_phaseC[:, 1:]
                cache_cirD[:, :-1] = cache_cirD[:, 1:]
                cache_amptiA[:, -1] = np.abs(csi[0, :, 0, 0])
                cache_amptiB[:, -1] = np.abs(csi[0, :, :, 0])
                cache_phaseC[:, -1] = calib(np.unwrap(np.angle(csi), axis=1), bw=BW, ng=NG)[0, :, :, 0]
                cache_cirD[:, -1] = np.abs(phy_ifft(csi[0, :, :, 0], axis=0, bw=BW, ng=NG))
                state = True
                mutex.unlock()
                count += 1
                if count % 100 == 0:
                    print('receive %d bytes [msgcnt=%u]' % (msg_len, count))


class MainWindow(QWidget):
    def __init__(self, parent=None, flags=Qt.WindowFlags()):
        super(MainWindow, self).__init__(parent=parent, flags=flags)
        self.subcarriers_index = scidx(bw=BW, ng=NG)
        self.configUI()
        self.configLayout()
        self.show()

    def configUI(self):
        self.setWindowTitle("CSIVIEWER")
        self.setWindowFlags(Qt.WindowCloseButtonHint | Qt.WindowMinimizeButtonHint | Qt.WindowMaximizeButtonHint)
        self.resize(800, 450)
        self.move(100, 100)
        self.setStyleSheet("border:0;background-color:#e8e8e8")

        # pygtgrapg global config
        pg.setConfigOptions(antialias=True)                 # Antialiasing
        pg.setConfigOptions(background=(232, 232, 232))     # White
        pg.setConfigOptions(foreground=(25, 25, 25))        # Black

        # plot area
        self.plot = pg.GraphicsWindow()
        self.initPlot()

        # ctrl area
        self.ctrl = QMenuBar(self)
        self.initCtrl()

        # update cache
        self.task = GetDataThread(self)
        self.task.start()

        # plot refresh
        self.PlotTimer()

    def configLayout(self):
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.plot)
        layout.addWidget(self.ctrl)

    @staticmethod
    def color_define(index):
        return (int(np.sin(index * (2 * np.pi) / 30 + 2) * 127 + 128),
                int(np.cos(index * (2 * np.pi) / 30 - 1) * 127 + 128),
                int(np.sin(index * (2 * np.pi) / 30 - 2) * 127 + 128),
                127)

    def initPlot(self):
        """Plot here"""
        # csi amplitude (time view) ======================================================
        pk_num = cache_amptiA.shape[1]
        p = self.plot.addPlot(title="Amplitude of CSI_800_%s_0_0" % (SUBCARRIERS_NUM))
        p.showGrid(x=False, y=True)
        p.setLabel('left', "Amplitude")
        p.setLabel('bottom', "Packets", units='')
        p.enableAutoRange('xy', False)
        p.setXRange(-int(pk_num/2), int(pk_num/2), padding=0.01)
        p.setYRange(0, YRange_A, padding=0.01)

        self.curvesA = []
        for i in range(SUBCARRIERS_NUM):
            color = self.color_define(i)
            self.curvesA.append(p.plot(pen=color, name='subcarrier=%d' % (i)))
        # x axis
        self.X = np.linspace(-int(pk_num/2), int(pk_num/2), pk_num)

        # csi amplitude (subcarriers view) ===============================================
        p2 = self.plot.addPlot(title="CFR of CSI_10_%s_3_0" % (SUBCARRIERS_NUM))
        p2.showGrid(x=False, y=True)
        p2.setLabel('left', "Amplitude")
        p2.setLabel('bottom', "Subcarriers", units='')
        p2.enableAutoRange('xy', False)
        p2.setXRange(-30, 30, padding=0.01)
        p2.setYRange(0, YRange_B, padding=0.01)

        self.curvesB = []
        for i in range(10*3):
            color = self.color_define(i)
            self.curvesB.append(p2.plot(pen=color))

        self.plot.nextRow()
        # csi phase (subcarriers view) ===================================================
        p3 = self.plot.addPlot(title="Phase of CSI_10_%s_3_0" % (SUBCARRIERS_NUM))
        p3.showGrid(x=False, y=True)
        p3.setLabel('left', "Phase")
        p3.setLabel('bottom', "Subcarriers", units='')
        p3.enableAutoRange('xy', False)
        p3.setXRange(-30, 30, padding=0.01)
        p3.setYRange(-np.pi, np.pi, padding=0.01)

        self.curvesC = []
        for i in range(10*3):
            color = self.color_define(i)
            self.curvesC.append(p3.plot(pen=color))

        # cir csi (time view) ============================================================
        p4 = self.plot.addPlot(title="CIR of CSI_10_%s_3_0" % (SUBCARRIERS_NUM))
        p4.showGrid(x=False, y=True)
        p4.setLabel('left', "Amplitude")
        p4.setLabel('bottom', "Time", units='50ns')
        p4.enableAutoRange('xy', False)
        p4.setXRange(0, 64, padding=0.01)
        p4.setYRange(0, YRange_D, padding=0.01)

        self.curvesD = []
        for i in range(10*3):
            color = self.color_define(i)
            self.curvesD.append(p4.plot(pen=color))

    def initCtrl(self):
        """Write your control here."""
        self.ctrl.addMenu("&File")
        self.ctrl.addMenu("&Edit")
        self.ctrl.addMenu("&Help")
        self.ctrl.setVisible(False)

    def PlotTimer(self):
        self.timer = QTimer()
        self.timer.setTimerType(0)
        self.timer.timeout.connect(self.updatePlot)
        self.timer.start(1000//30)

    @pyqtSlot()
    def updatePlot(self):
        global cache_amptiA, cache_amptiB, cache_phaseC, cache_cirD, mutex, state
        mutex.lock()
        if state:
            for i in range(SUBCARRIERS_NUM):
                self.curvesA[i].setData(self.X, cache_amptiA[i])
            for i in range(10*3):
                self.curvesB[i].setData(self.subcarriers_index, cache_amptiB[:, i%10 , i//10])
                self.curvesC[i].setData(self.subcarriers_index, cache_phaseC[:, i%10 , i//10])
                self.curvesD[i].setData(np.arange(64), cache_cirD[:, i%10 , i//10])
            state = False
        mutex.unlock()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_F11:
            self.setWindowState(self.windowState() ^ Qt.WindowFullScreen)
        if event.key() == Qt.Key_F1:
            self.ctrl.setVisible(not self.ctrl.isVisible())
        event.accept()

    def closeEvent(self, event):
        self.task.stop()
        event.accept()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = MainWindow()
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Parse MPDU

     2          2          6        6         6         2          6         2         4      0-7955   4  
| Frame   | Duration | Address | Address | Address | Sequence | Address |   Qos   |   HT    | Frame | FCS |
| Control |   /ID    |    1    |    2    |    3    |  Control |    4    | Control | Control |  Body |     |
<-----------------------------------------MAC Header---------------------------------------->

The first three fields (Frame Control, Duration/ID, and Address 1) and the last field (FCS) 
constitute the minimal frame format and are present in all frames, including reserved types
and subtypes. The fields Address 2, Address 3, Sequence Control, Address 4, QoS Control, 
HT Control, and Frame Body are present only in **certain frame types and subtypes**.

------------------------------------------------------------------------------------------

csiread.Intel only parses MPDU defined in `random_packets.c` and csiread.Atheros does nothing
to MPDU. (MPDU defined in `random_packets.c` is not network byte order)

ref:
    1. [CWAP – MAC Header : Frame Control](https://mrncciew.com/2014/09/27/cwap-mac-header-frame-control/)
    2. IEEE 802.11n Standard
"""

import struct

import csiread
import numpy as np
from scapy.all import Dot11


class MPDU:
    def __init__(self, frame, endian='!'):
        if len(frame) < 40:
            raise Exception('Error: pl_size >= 40 is requried')
        self.frame_control, = struct.unpack(endian + 'H', frame[:2])
        self.duration, = struct.unpack(endian + 'H', frame[2:4])
        self.addr1, = struct.unpack(endian + '6s', frame[4:10])
        self.addr2, = struct.unpack(endian + '6s', frame[10:16])
        self.addr3, = struct.unpack(endian + '6s', frame[16:22])
        self.sequence_control, = struct.unpack(endian + 'H', frame[22:24])
        self.addr4, = struct.unpack(endian + '6s', frame[24:30])
        self.qos_control, = struct.unpack(endian + 'H', frame[30:32])
        self.ht_control, = struct.unpack(endian + 'L', frame[32:36])
        self.frame_body, = struct.unpack(endian + '4s', frame[36:40])

    def show(self):
        t = "{:20s}{:>12} {:<24s}\n"

        s = "-"*50 + '\n'
        s += t.format('Field', 'Hex', 'Value')
        s += "-"*50 + '\n'
        s += t.format('frame_control: ', hex(self.frame_control)[2:].zfill(4), str(self.frame_control))
        s += t.format('duration: ', hex(self.duration)[2:].zfill(4), str(self.duration))
        s += t.format('address_1: ', self.func(self.addr1, ''),  self.func(self.addr1, ':'))
        s += t.format('address_2: ',  self.func(self.addr2, ''),  self.func(self.addr2, ':'))
        s += t.format('address_3: ',  self.func(self.addr3, ''),  self.func(self.addr3, ':'))
        s += t.format('sequence_control: ', hex(self.sequence_control)[2:].zfill(4), str(self.sequence_control))
        s += t.format('address_4: ',  self.func(self.addr4, ''),  self.func(self.addr4, ':'))
        s += t.format('qos_control: ', hex(self.qos_control)[2:].zfill(4), str(self.qos_control))
        s += t.format('ht_control: ', hex(self.ht_control)[2:].zfill(8), str(self.ht_control))
        s += t.format('frame_body[:4]: ',  self.func(self.frame_body[:4], ''),  self.func(self.frame_body[:4], ' '))
        s += "-"*50
        print(s)

    @staticmethod
    def func(vlist, space):
        return space.join([hex(per)[2:].zfill(2) for per in vlist])

def AtherosMPDU(csifile, index=0):
    """Parse payload using scapy"""
    csidata = csiread.Atheros(csifile, nrxnum=3, ntxnum=2, pl_size=40, tones=56, if_report=False)
    csidata.read()
    print("=" * 20 + "Atheros MPDU" + "=" * 20)
    m = Dot11(csidata.payload[index].tobytes())
    m.show()

def Intel5300MPDU(csifile, index=0):
    """Parse payload using scapy"""
    csidata = csiread.Intel(csifile, nrxnum=3, ntxnum=2, pl_size=40, if_report=False)
    csidata.read()
    print("=" * 20 + "Intel 5300 MPDU" + "=" * 20)
    m = Dot11(csidata.payload[index].tobytes())
    m.show()

def AtherosMPDU2(csifile, index=0):
    """Parse payload without scapy"""
    csidata = csiread.Atheros(csifile, nrxnum=3, ntxnum=2, pl_size=40, tones=56, if_report=False)
    csidata.read()
    print("=" * 20 + "Atheros MPDU2" + "=" * 20)
    m = MPDU(csidata.payload[index].tobytes())
    m.show()

def Intel5300MPDU2(csifile, index=0):
    """Parse payload without scapy"""
    csidata = csiread.Intel(csifile, nrxnum=3, ntxnum=2, pl_size=40, if_report=False)
    csidata.read()
    print("=" * 20 + "Intel 5300 MPDU2" + "=" * 20)
    m = MPDU(csidata.payload[index].tobytes())
    m.show()


if __name__ == "__main__":
    AtherosMPDU("../material/atheros/dataset/ath_csi_1.dat", 10)
    Intel5300MPDU("../material/5300/dataset/sample_0x5_64_3000.dat", 10)



"""Check correctness(Linux 802.11n CSI Tool)

Usage:
	1. use the following matlab code to get 'log.all_csi.6.7.6.mat'.

	```matlab
	csidata = read_bf_file("sample_data/log.all_csi.6.7.6");
	for i = 1:length(csidata)
		csidata{i}.scaled_csi = get_scaled_csi(csidata{i});
		csidata{i}.scaled_csi_sm = get_scaled_csi_sm(csidata{i});
	end
	save("sample_data/log.all_csi.6.7.6.mat", 'csidata')
	```

	2. python3 csitest.py

Note:
	the sample data 'log.all_csi.6.7.6' of linux-80211n-csitool-supplementary
	contains Ntx=1, 2, 3.
"""

import numpy as np
from scipy.io import loadmat
import csiread


def python_read(csifile='log.all_csi.6.7.6'):
	csidata = csiread.Intel(csifile, ntxnum=3)
	csidata.read()
	pydata = csidata[:]
	pydata['scaled_csi'] = csidata.get_scaled_csi()
	pydata['scaled_csi_sm'] = csidata.get_scaled_csi_sm()
	return pydata


def matlab_read(matfile='log.all_csi.6.7.6.mat'):
	matdata = loadmat(matfile)
	mldata = {}
	pk_num = 29
	mldata['timestamp_low'] = np.zeros([pk_num], dtype=np.int_)
	mldata['bfee_count'] = np.zeros([pk_num], dtype=np.int_)
	mldata['Nrx'] = np.zeros([pk_num], dtype=np.int_)
	mldata['Ntx'] = np.zeros([pk_num], dtype=np.int_)
	mldata['rssi_a'] = np.zeros([pk_num], dtype=np.int_)
	mldata['rssi_b'] = np.zeros([pk_num], dtype=np.int_)
	mldata['rssi_c'] = np.zeros([pk_num], dtype=np.int_)
	mldata['noise'] = np.zeros([pk_num], dtype=np.int_)
	mldata['agc'] = np.zeros([pk_num], dtype=np.int_)
	mldata['perm'] = np.zeros([pk_num, 3], dtype=np.int_)
	mldata['rate'] = np.zeros([pk_num], dtype=np.int_)
	mldata['csi'] = np.zeros([pk_num, 30, 3, 3], dtype=np.complex_)
	mldata['scaled_csi'] = np.zeros([pk_num, 30, 3, 3], dtype=np.complex_)
	mldata['scaled_csi_sm'] = np.zeros([pk_num, 30, 3, 3], dtype=np.complex_)
	for i in range(pk_num):
		mldata['timestamp_low'][i] = matdata['csidata'][i, 0]['timestamp_low'][0, 0]
		mldata['bfee_count'][i] = matdata['csidata'][i, 0]['bfee_count'][0, 0]
		mldata['Nrx'][i] = matdata['csidata'][i, 0]['Nrx'][0, 0]
		mldata['Ntx'][i] = matdata['csidata'][i, 0]['Ntx'][0, 0]
		mldata['rssi_a'][i] = matdata['csidata'][i, 0]['rssi_a'][0, 0]
		mldata['rssi_b'][i] = matdata['csidata'][i, 0]['rssi_b'][0, 0]
		mldata['rssi_c'][i] = matdata['csidata'][i, 0]['rssi_c'][0, 0]
		mldata['noise'][i] = matdata['csidata'][i, 0]['noise'][0, 0]
		mldata['agc'][i] = matdata['csidata'][i, 0]['agc'][0, 0]
		mldata['perm'][i] = matdata['csidata'][i, 0]['perm'][0, 0].T[:, 0] - 1
		mldata['rate'][i] = matdata['csidata'][i, 0]['rate'][0, 0]
		mldata['csi'][i][:, :, :int(mldata['Ntx'][i])] = matdata['csidata'][i, 0]['csi'][0, 0].T
		mldata['scaled_csi'][i][:, :, :int(mldata['Ntx'][i])] = matdata['csidata'][i, 0]['scaled_csi'][0, 0].T
		mldata['scaled_csi_sm'][i][:, :, :int(mldata['Ntx'][i])] = matdata['csidata'][i, 0]['scaled_csi_sm'][0, 0].T
	return mldata


if __name__ == "__main__":
	pydata = python_read('log.all_csi.6.7.6')
	mldata = matlab_read('log.all_csi.6.7.6.mat')

	for i in mldata.keys():
		print('%-15s' % (i), np.sum(np.abs(pydata[i] - mldata[i])))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Calculate effective SNR based on CSI(Linux 802.11n CSI Tool)

Ref: [Predictable 802.11 Packet Delivery from Wireless Channel Measurements]
    (https://www.halper.in/pubs/comm356s-halperin.pdf) and [linux-80211n-csitool-supplementary]
    (https://github.com/dhalperi/linux-80211n-csitool-supplementary)
"""


import csiread
import matplotlib.pyplot as plt
import numpy as np
from numpy.linalg import inv
from scipy.special import erfc, erfcinv


def dbinv(x):
    return np.power(10, x/10)


def db(x):
    return 10 * np.log10(x)


def qfunc(x):
    return 0.5 * erfc(x/np.sqrt(2))


def qfuncinv(x):
    return np.sqrt(2) * erfcinv(2*x)


def bpsk_ber(snr):
    return qfunc(np.sqrt(2*snr))


def bpsk_berinv(ber):
    return np.power(qfuncinv(ber), 2.0) / 2


def qam16_ber(snr):
    return 3/4 * qfunc(np.sqrt(snr/5))


def qam16_berinv(ber):
    return np.power(qfuncinv(ber * 4/3), 2.0) * 5


def qam64_ber(snr):
    return 7/12*qfunc(np.sqrt(snr/21))


def qam64_berinv(ber):
    return np.power(qfuncinv(ber * 12/7), 2.0) * 21


def qpsk_ber(snr):
    return qfunc(np.sqrt(snr))


def qpsk_berinv(ber):
    return np.power(qfuncinv(ber), 2.0)


def mimo2_mmse(csi):
    """csi.shape = (Ntx(2), Nrx)"""
    M = inv(np.dot(csi.T.conjugate(), csi) + np.eye(2))
    ret = 1 / np.diag(M) - 1
    return ret.real


def mimo3_mmse(csi):
    """csi.shape = (Ntx(3), Nrx)"""
    M = inv(np.dot(csi.T.conjugate(), csi) + np.eye(3))
    ret = 1 / np.diag(M) - 1
    return ret.real


def get_simo_SNRs(csi):

    return np.sum(csi * csi.conjugate(), axis=(1), keepdims=True).real


def get_mimo2_SNRs(csi):
    S, N, M = csi.shape
    csi = csi / np.sqrt(2.0)
    if M == 2:
        ret = np.zeros([S, 2, 1])
        for i in range(S):
            ret[i, :, 0] = mimo2_mmse(csi[i])
        return ret

    ret = np.zeros([S, 2, 3])
    for i in range(S):
        ret[i, :, 0] = mimo2_mmse(csi[i, :, 0:2])
        ret[i, :, 1] = mimo2_mmse(csi[i, :, 0:3:2])
        ret[i, :, 2] = mimo2_mmse(csi[i, :, 1:3])
    return ret


def get_mimo3_SNRs(csi):
    S, N, M = csi.shape
    csi = csi / np.sqrt(dbinv(4.5))

    ret = np.zeros([S, 3, 1])
    for i in range(S):
        ret[i, :, 0] = mimo3_mmse(csi[i])
    return ret


def get_eff_SNRs(csi, csi_sm=None):
    """get_eff_SNRs

    When csi_sm is given, this function is get_eff_SNRs_sm.
    """
    if len(csi.shape) == 3:
        csi = np.array([csi])

    L, S, N, M = csi.shape
    ret = np.zeros([L, 7, 4])

    k = min(M, N)
    if k >= 1:
        for j in range(L):
            snrs = get_simo_SNRs(csi[j])
            bers = bpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, :len(mean_ber), 0] = bpsk_berinv(mean_ber)

            bers = qpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, :len(mean_ber), 1] = qpsk_berinv(mean_ber)

            bers = qam16_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, :len(mean_ber), 2] = qam16_berinv(mean_ber)

            bers = qam64_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, :len(mean_ber), 3] = qam64_berinv(mean_ber)

    if csi_sm is not None:
        csi = np.array([csi_sm]) if len(csi_sm.shape) == 3 else csi_sm

    if k >= 2:
        for j in range(L):
            snrs = get_mimo2_SNRs(csi[j])

            bers = bpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 3:3+len(mean_ber), 0] = bpsk_berinv(mean_ber)

            bers = qpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 3:3+len(mean_ber), 1] = qpsk_berinv(mean_ber)

            bers = qam16_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 3:3+len(mean_ber), 2] = qam16_berinv(mean_ber)

            bers = qam64_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 3:3+len(mean_ber), 3] = qam64_berinv(mean_ber)

    if k >= 3:
        for j in range(L):
            snrs = get_mimo3_SNRs(csi[j])

            bers = bpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 6:6+len(mean_ber), 0] = bpsk_berinv(mean_ber)

            bers = qpsk_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 6:6+len(mean_ber), 1] = qpsk_berinv(mean_ber)

            bers = qam16_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 6:6+len(mean_ber), 2] = qam16_berinv(mean_ber)

            bers = qam64_ber(snrs)
            mean_ber = np.mean(bers, axis=(0, 1))
            ret[j, 6:6+len(mean_ber), 3] = qam64_berinv(mean_ber)
    ret[ret == np.Inf] = dbinv(40)
    return ret


def main():
    csidata = csiread.Intel('../material/5300/dataset/sample_0x5_64_3000.dat')
    csidata.read()
    scaled_csi = csidata.get_scaled_csi()
    preNrx = max(csidata.Nrx)
    preNtx = max(csidata.Ntx)
    if max(csidata.Nrx) != min(csidata.Nrx):
        print('be careful')
        preNrx = 3  # set by yourself
        scaled_csi = scaled_csi[csidata.Nrx == preNrx]
    if max(csidata.Ntx) != min(csidata.Ntx):
        print('be careful')
        preNtx = 2  # set by yourself
        scaled_csi = scaled_csi[csidata.Ntx == preNtx]
    scaled_csi = scaled_csi[:100, :, :preNrx, :preNtx]
    ret = get_eff_SNRs(scaled_csi)
    print(ret[0])
    plt.figure()
    plt.plot(ret[:, 0, :])
    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from setuptools import distutils
from setuptools.extension import Extension
from Cython.Build import cythonize
import numpy

default_compiler = distutils.ccompiler.get_default_compiler()
if default_compiler == 'unix':
    EXTRA_COMPILE_ARGS = ['-g0']
else:
    EXTRA_COMPILE_ARGS = []

LONG_DESCRIPTION = """\
# csiread [![PyPI](https://img.shields.io/pypi/v/csiread?)](https://pypi.org/project/csiread/)

A fast channel state information parser for Intel, Atheros and Nexmon in Python.

- Full support for Linux 802.11n CSI Tool, Atheros CSI Tool and nexmon_csi
- At least 15 times faster than the implementation in Matlab
- Real-time parsing and visualization.
"""

EXTENSIONS = [
    Extension(
        "csiread.csiread", ["csiread/csiread.pyx"],
        include_dirs=[numpy.get_include()],
        define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
        extra_compile_args=EXTRA_COMPILE_ARGS,
    ),
]

setup(
    name="csiread",
    version="1.3.6",

    description="A fast channel state information parser for Intel, " \
                "Atheros and Nexmon.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",

    author="Hecheng Su",
    author_email="2215523266@qq.com",
    url="https://github.com/citysu/csiread",

    packages=find_packages(),
    install_requires=['numpy'],
    python_requires='>=3',
    ext_modules=cythonize(
        EXTENSIONS,
        compiler_directives={'language_level': 3, 'binding': False}
    ),

    license='MIT',
    classifiers=["Topic :: Scientific/Engineering",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: Implementation :: CPython",
                 "Operating System :: Unix",
                 "Operating System :: POSIX",
                 "Operating System :: Microsoft",
                 "Operating System :: MacOS",
                 "License :: OSI Approved :: MIT License"],
)



"""How to use seek() method

Usage:
    python3 csiseek.py

Note:
    1. This example should be tested with large files.
    2. ...
"""
import csiread
import os
import random
from timeit import default_timer
from utils import infer_device


def getPos(csifile):
    device = infer_device(csifile)
    if device == 'Intel':
        return getPosIntel(csifile)
    elif device == 'Atheros':
        return getPosAtheros(csifile)
    else:
        return getPosNexmon(csifile)


def getPosIntel(csifile):
    lens = os.path.getsize(csifile)
    cur, positions = 0, []
    last = [None, None]
    with open(csifile, 'rb') as f:
        while cur < (lens - 3):
            field_len = int.from_bytes(f.read(2), byteorder='big')
            code = int.from_bytes(f.read(1), byteorder='little')
            if code == 0xbb:
                if last[0] == 0xc1:
                    positions.append(cur - last[1] - 2)
                else:
                    positions.append(cur)
            last = [code, field_len]
            cur += (field_len + 2)
            f.seek(field_len - 1, os.SEEK_CUR)
    return positions


def getPosAtheros(csifile):
    lens = os.path.getsize(csifile)
    cur, positions = 0, []
    with open(csifile, 'rb') as f:
        while cur < (lens - 4):
            field_len = int.from_bytes(f.read(2), byteorder='little')
            f.seek(field_len, os.SEEK_CUR)
            positions.append(cur)
            cur += (field_len + 2)
    return positions


def getPosNexmon(csifile):
    lens = os.path.getsize(csifile)
    cur, positions = 24, []
    with open(csifile, 'rb') as f:
        magic = f.read(4)
        f.seek(20, os.SEEK_CUR)
        if magic in [b"\xa1\xb2\xc3\xd4", b"\xa1\xb2\x3c\x4d"]:
            endian = 'big'
        else:
            endian = 'little'
        while cur < (lens - 4):
            caplen = int.from_bytes(f.read(16)[8:12], byteorder=endian)
            if f.read(42)[6:12] == b'NEXMON':
                positions.append(cur)
            f.seek(caplen - 42, os.SEEK_CUR)
            cur += (caplen + 16)
    return positions


def seekIntel(csifile, num=36):
    positions_all = getPos(csifile)
    positions = positions_all[::num][:-1]
    random.shuffle(positions)
    print("-"*40, "[Intel]")

    last = default_timer()
    csidata = csiread.Intel(csifile, 3, 2, 0, False)
    csidata.read()
    cost_time = default_timer() - last
    print("%-18s: %18fs" % ("read()", cost_time))

    last = default_timer()
    csidata = csiread.Intel(None, 3, 2, 0, False, bufsize=num)
    for pos in positions:
        csidata.seek(csifile, pos, num)
    cost_time = (default_timer() - last) * len(positions_all) / (len(positions) * num)
    print("%-18s: %18fs" % ("seek(num=%d)" % (num), cost_time))


def seekAtheros(csifile, num=36):
    positions_all = getPos(csifile)
    positions = positions_all[::num][:-1]
    random.shuffle(positions)
    print("-"*40, "[Atheros]")

    last = default_timer()
    csidata = csiread.Atheros(csifile, 3, 2, 0, 56, False)
    csidata.read()
    cost_time = default_timer() - last
    print("%-18s: %18fs" % ("read()", cost_time))

    last = default_timer()
    csidata = csiread.Atheros(None, 3, 2, 0, 56, False, bufsize=num)
    for pos in positions:
        csidata.seek(csifile, pos, num)
    cost_time = (default_timer() - last) * len(positions_all) / (len(positions) * num)
    print("%-18s: %18fs" % ("seek(num=%d)" % (num), cost_time))


def seekNexmon(csifile, num=36):
    positions_all = getPos(csifile)
    positions = positions_all[::num][:-1]
    random.shuffle(positions)
    print("-"*40, "[Nexmon]")

    last = default_timer()
    csidata = csiread.Nexmon(csifile, '4358', 80, if_report=False)
    csidata.read()
    cost_time = default_timer() - last
    print("%-18s: %18fs" % ("read()", cost_time))

    last = default_timer()
    csidata = csiread.Nexmon(None, '4358', 80, if_report=False, bufsize=num)
    for pos in positions:
        csidata.seek(csifile, pos, num)
    cost_time = (default_timer() - last) * len(positions_all) / (len(positions) * num)
    print("%-18s: %18fs" % ("seek(num=%d)" % (num), cost_time))


if __name__ == "__main__":
    csifile_intel = "../material/5300/dataset/sample_0x5_64_3000.dat"
    csifile_atheros = "../material/atheros/dataset/ath_csi_1.dat"
    csifile_nexmon = "../material/nexmon/dataset/example.pcap"
    seekIntel(csifile_intel, 64)
    seekAtheros(csifile_atheros, 64)
    seekNexmon(csifile_nexmon, 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Plot csi of Linux 802.11n CSI Tool, observe csi quickly

Usage:
    python3 csishow.py ../material/5300/dataset/sample_0x1_ap.dat -t 1
"""
import argparse

import csiread
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal
from utils import scidx, calib, phy_ifft


def func_1(csidata):
    """CSI: time-amplitude"""
    s_index = 15    # subcarrier index
    csi = csidata.get_scaled_csi_sm()
    t = csidata.timestamp_low/1000000 - csidata.timestamp_low[0]/1000000

    amplitude = np.abs(csi[:, s_index])

    plt.figure()
    plt.plot(t, amplitude[:, 0, 0], linewidth=0.3, label='subcarrier_15_0_0')
    plt.plot(t, amplitude[:, 1, 0], linewidth=0.3, label='subcarrier_15_1_0')
    plt.plot(t, amplitude[:, 2, 0], linewidth=0.3, label='subcarrier_15_2_0')
    plt.legend()

    plt.title('csi-amplitude')
    plt.xlabel('time(s)')
    plt.ylabel('amplitude')
    plt.show()


def func_2(csidata):
    """CSI: subcarrier-amplitude(CFR)"""
    csi = csidata.get_scaled_csi_sm()
    amplitude = np.abs(csi)
    s_index = scidx(20, 2)

    plt.figure()
    plt.plot(s_index, amplitude[:100, :, 0, 0].T, 'r-', linewidth=0.3)
    plt.plot(s_index, amplitude[:100, :, 1, 0].T, 'g-', linewidth=0.3)
    plt.plot(s_index, amplitude[:100, :, 2, 0].T, 'y-', linewidth=0.3)

    patch_1 = mpatches.Patch(color='red', label=':100_r0t0')
    patch_2 = mpatches.Patch(color='green', label=':100_r1t0')
    patch_3 = mpatches.Patch(color='yellow', label=':100_r2t0')
    plt.legend(handles=[patch_1, patch_2, patch_3])

    plt.title('csi-amplitude')
    plt.xlabel('subcarriers')
    plt.ylabel('amplitude')
    plt.show()


def func_3(csidata):
    """CSI: time-phase"""
    s_index = 15    # subcarrier index
    csi = csidata.get_scaled_csi_sm()
    t = csidata.timestamp_low/1000000 - csidata.timestamp_low[0]/1000000
    phase = np.unwrap(np.angle(csi), axis=1)
    phase = calib(phase)

    plt.figure()
    plt.plot(t, phase[:, s_index, 0, 0], linewidth=0.3, label='subcarrier_15_0_0')
    plt.plot(t, phase[:, s_index, 1, 0], linewidth=0.3, label='subcarrier_15_1_0')
    plt.plot(t, phase[:, s_index, 2, 0], linewidth=0.3, label='subcarrier_15_2_0')
    plt.legend()
    plt.title('csi-phase')
    plt.xlabel('time(s)')
    plt.ylabel('phase')
    plt.show()


def func_4(csidata):
    """CSI: subcarrier-phase"""
    csi = csidata.get_scaled_csi_sm()
    phase = np.unwrap(np.angle(csi), axis=1)
    phase = calib(phase)
    s_index = scidx(20, 2)

    plt.figure(4)
    plt.plot(s_index, phase[:100, :, 0, 0].T, 'r-', linewidth=0.3)
    plt.plot(s_index, phase[:100, :, 1, 0].T, 'g-', linewidth=0.3)
    plt.plot(s_index, phase[:100, :, 2, 0].T, 'y-', linewidth=0.3)

    patch_1 = mpatches.Patch(color='red', label=':100_r0t0')
    patch_2 = mpatches.Patch(color='green', label=':100_r1t0')
    patch_3 = mpatches.Patch(color='yellow', label=':100_r2t0')
    plt.legend(handles=[patch_1, patch_2, patch_3])

    plt.title('csi-phase')
    plt.xlabel('subcarriers')
    plt.ylabel('phase')
    plt.show()


def func_5(csidata):
    """timestamp_low: packet-time_difference"""
    time_diff = np.diff(csidata.timestamp_low)
    plt.figure(5)
    plt.plot(time_diff, linewidth=0.3, label='time diff')
    plt.legend()
    plt.title('time-diff')
    plt.xlabel('packets')
    plt.ylabel('time(us)')
    plt.show()


def func_6(csidata):
    """CSI: time-amplitude(CIR: OFDM symbol view)"""
    csi = csidata.get_scaled_csi_sm()

    amplitude1 = np.abs(phy_ifft(csi[:100, :, 0, 0], axis=1)).T
    amplitude2 = np.abs(phy_ifft(csi[:100, :, 1, 0], axis=1)).T
    amplitude3 = np.abs(phy_ifft(csi[:100, :, 2, 0], axis=1)).T
    t = np.linspace(0, 64, 64)

    plt.figure(6)
    plt.plot(t, amplitude1, 'r-', linewidth=0.3)
    plt.plot(t, amplitude2, 'g-', linewidth=0.3)
    plt.plot(t, amplitude3, 'y-', linewidth=0.3)

    patch_1 = mpatches.Patch(color='red', label=':100_r0t0')
    patch_2 = mpatches.Patch(color='green', label=':100_r1t0')
    patch_3 = mpatches.Patch(color='yellow', label=':100_r2t0')
    plt.legend(handles=[patch_1, patch_2, patch_3])

    plt.title('csi-CIR')
    plt.xlabel('time(50ns)')
    plt.ylabel('amplitude')
    plt.show()


def func_8(csidata):
    """CSI: time-amplitude(butterworth filter)"""
    print("Waring: Please set parameters(butterworth) first")

    csi = csidata.get_scaled_csi_sm()[:1000]
    amplitude = np.abs(csi)
    t = (csidata.timestamp_low/1000000 - csidata.timestamp_low[0]/1000000)[:1000]

    sos = signal.butter(4, 100.0, 'lowpass', fs=900, output='sos')
    sfiltered = signal.sosfiltfilt(sos, amplitude, axis=0)

    plt.figure(9)
    plt.plot(t, amplitude[:, 15, 0, 0], '-', linewidth=1.5, alpha=0.25, label='origin_:1000s15r0t0')
    plt.plot(t, amplitude[:, 15, 1, 0], '-', linewidth=1.5, alpha=0.25, label='origin_:1000s15r1t0')
    plt.plot(t, amplitude[:, 15, 2, 0], '-', linewidth=1.5, alpha=0.25, label='origin_:1000s15r2t0')

    plt.plot(t, sfiltered[:, 15, 0, 0], '-', linewidth=1, label='buffer_:1000s15r0t0')
    plt.plot(t, sfiltered[:, 15, 1, 0], '-', linewidth=1, label='buffer_:1000s15r1t0')
    plt.plot(t, sfiltered[:, 15, 2, 0], '-', linewidth=1, label='buffer_:1000s15r2t0')

    plt.legend()
    plt.title("butterworth - lowpass")
    plt.xlabel('time(s)')
    plt.ylabel('amplitude')
    plt.show()


def func_9(csidata):
    index = 15
    csi = csidata.get_scaled_csi_sm()
    amplitude = np.abs(csi)
    t = (csidata.timestamp_low/1000000 - csidata.timestamp_low[0]/1000000)[:1000]

    kernel = np.array([[1], [2], [3], [2], [1]]) / 9
    amplitude1 = signal.convolve(amplitude[:1000, index:index+1, 0, 0], kernel, mode='same')
    amplitude2 = signal.convolve(amplitude[:1000, index:index+1, 1, 0], kernel, mode='same')
    amplitude3 = signal.convolve(amplitude[:1000, index:index+1, 2, 0], kernel, mode='same')

    plt.figure()
    plt.plot(t, amplitude1, 'r-', linewidth=0.3, label=':_s15r0t0')
    plt.plot(t, amplitude2, 'g-', linewidth=0.3, label=':_s15r1t0')
    plt.plot(t, amplitude3, 'y-', linewidth=0.3, label=':_s15r2t0')

    plt.legend()
    plt.title('csi amplitude conv smooth')
    plt.xlabel('time(s)')
    plt.ylabel('amplitude')
    plt.show()


helpinfo = r"""Plot Type

1   CSI-time-amplitude
2   CSI-subcarrier-amplitude
3   CSI-time-phase
4   CSI-subcarrier-phase
5   timestamp-packet-timediff
6   CSI-time-amplitude(CIR: OFDM symbol view)
8   CSI-time-amplitude(butterworth filter)
9   CSI-time-amplitude(Convolve)
10  CSI-stft
"""


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('file', type=str, help='data file')
    parser.add_argument('-t', type=int, default=1, help=helpinfo)
    p = parser.parse_args()

    csidata = csiread.Intel(p.file, nrxnum=3, ntxnum=3)
    csidata.read()
    if p.t > 10:
        raise ValueError('the value of `t` can be 1 - 10')
    func = eval('func_' + str(p.t))
    func(csidata)
